"""
Chainer code (version 1.23.0) for DCNN saliency map model
Class for DCNN saliency map model is shown here.

class DeepNetSaliency:
Architecture of Pan Model

class CNN:
including the functions for training and testing the model.

"""

import time
import six.moves.cPickle as pickle
import numpy as np
from sklearn.datasets import fetch_mldata
from sklearn.cross_validation import train_test_split
from chainer import cuda, Variable, FunctionSet, optimizers, serializers
import chainer.functions as F
import chainer.links as L # for test
import test_functions as tf
import copy
from saliency import SalientResult
from scipy import io, sparse

class DeepNetSaliency(FunctionSet): #class chainer.FunctionSet
    def __init__(self):
        super(DeepNetSaliency, self).__init__(
            conv1=  L.Convolution2D(3, 96, 7, stride=1, pad=3),
            conv2=  L.Convolution2D(96, 256, 5, stride=1, pad=2),
            conv3=  L.Convolution2D(256, 512, 3, stride=1, pad=1),
            conv4=  L.Convolution2D(512, 512, 5, stride=1, pad=2),
            conv5=  L.Convolution2D(512, 512, 5, stride=1, pad=2),
            conv6=  L.Convolution2D(512, 256, 7, stride=1, pad=3),
            conv7=  L.Convolution2D(256, 128, 11, stride=1, pad=5),
            conv8=  L.Convolution2D(128, 32, 11, stride=1, pad=5),
            conv9=  L.Convolution2D(32, 1, 13, stride=1, pad=6),
            deconv= L.Deconvolution2D(1, 1, ksize=8, stride=4, pad=2)
        )

    def forward(self, x_data, y_data, train=True, gpu=-1):
        # x_data : data for train data (size : batch size)
        # y_data : target for train data  (size : batch size)

        if gpu >= 0:
            x_data = cuda.to_gpu(x_data)
            y_data = cuda.to_gpu(y_data)

        x, t = Variable(x_data), Variable(y_data) # x:chainer.variable.Variable
        
        h = F.max_pooling_2d(F.local_response_normalization(F.relu(self.conv1(x))), ksize=3, stride=2)
        h = F.max_pooling_2d(F.relu(self.conv2(h)), ksize=3, stride=2)
        h = F.relu(self.conv3(h))
        h = F.relu(self.conv4(h))
        h = F.relu(self.conv5(h))
        h = F.relu(self.conv6(h))
        h = F.relu(self.conv7(h))
        h = F.relu(self.conv8(h))
        h = self.conv9(h)
        y = self.deconv(h)
        
        return y, F.mean_squared_error(y, t)


class CNN: #cnn.x_train,cnn.n_train <- self.params self.structure 
    def __init__(self, data, gpu=-1):

        self.model = DeepNetSaliency()
        self.model_name = 'cnn_model'

        if gpu >= 0:
            self.model.to_gpu()

        self.gpu = gpu

        self.data_train = data #file name list
        self.dir_data = '../TestData/'
        self.dir_sal = '../GroundTruthFG/'
        self.n_data = len(self.data_train)

        self.optimizer = optimizers.Adam(alpha=0.00005)
        self.optimizer.setup(self.model.collect_parameters())

    def predict(self, x_data, gpu=-1):
        return self.model.predict(x_data, gpu)


    # for Training the DCNN saliency map model
    def train(self, n_epoch=100, batchsize=100): 
        epoch = 1
        best_accuracy = 0
        list_epoch,list_loss = [],[]
        while epoch <= n_epoch:
            print 'epoch', epoch

            perm = np.random.permutation(self.n_data) 
            sum_train_loss = 0
            
            for i in xrange(0, self.n_data, batchsize): 
                data_list = []
                for j in perm[i:i+batchsize]:
                    data_list.append(self.data_train[j])
                data_batch,sal_batch,imgname_batch,salname_batch = tf.load_test_batch(data_list,self.dir_data,self.dir_sal) #for testing
                
                real_batchsize = len(data_batch) 
                self.optimizer.zero_grads() # Fills all gradients array by zeros 
                sal, loss = self.model.forward(data_batch, sal_batch, train=True, gpu=self.gpu)
                loss.backward() # send the error backpropagation 
                self.optimizer.update() 

                loss_batch = float(cuda.to_cpu(loss.data)) * real_batchsize
                sum_train_loss += loss_batch

            print 'train mean loss={}'.format(sum_train_loss/self.n_data)
            
            list_epoch.append(epoch)
            list_loss.append(sum_train_loss/self.n_data)
            if epoch % 10 == 0:
                serializers.save_npz('./Model/ModelEpoch'+str(epoch),copy.deepcopy(self.model).to_cpu())
            epoch += 1
            # for testing model
            print "Testing the Model"
            sal_result = []
            sal, loss = self.model.forward(data_batch, sal_batch, train=False, gpu=self.gpu) 
            for i_batch in range(0,real_batchsize):
                sal_result.append(SalientResult(imgname_batch[i_batch]))
                sal.to_cpu()
                sal_result[len(sal_result)-1].data = sal[i_batch][0].data
                tf.save_saliency(sal_result,'./testResults/testResultEpoch'+str(epoch-1))
        return list_loss, list_epoch

    #for testing the model
    def test(self, img, groundtruth, dir_dataname, dir_salname, batchsize=100):
        """
        img : list of img file name
        groundtruth : list of groundtruth file name
        """
        sal_result = []
        num_img = len(img)
        for i in xrange(0, num_img, batchsize):
            data_list = []
            for j in img[i:i+batchsize]:
                data_list.append(j)
            data_batch,sal_batch,imgname_batch,salname_batch = tf.load_test_batch(data_list,dir_dataname,dir_salname)
            real_batchsize = len(data_batch) 
            sal, loss = self.model.forward(data_batch, sal_batch, train=False, gpu=self.gpu)
            for i_batch in range(0,real_batchsize):
                sal_result.append(SalientResult(imgname_batch[i_batch]))
                sal.to_cpu()
                sal_result[len(sal_result)-1].data = sal[i_batch][0].data
        return sal_result

    def dump_model(self):
        self.model.to_cpu()
        pickle.dump(self.model, open(self.model_name, 'wb'), -1)

    def load_model(self):
        self.model = pickle.load(open(self.model_name,'rb'))
        if self.gpu >= 0:
            self.model.to_gpu()
        self.optimizer.setup(self.model.collect_parameters())
